-------------------------------------------------------------------------------
-- Title      : AD9512 clock divider synchronizer
-- Project    : FMC DEL 1ns 2cha (FFPG)
-- URL        : http://www.ohwr.org/projects/fmc-del-1ns-2cha
-------------------------------------------------------------------------------
-- File       : Ad9512Syncer.vhd
-- Author(s)  : Jan Pospisil <j.pospisil@cern.ch>
-- Company    : CERN (BE-BI-QP)
-- Created    : 2016-08-17
-- Last update: 2016-08-24
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: On request, generates AD9512 synchronization pulse based on
--              Trigger_i.
-------------------------------------------------------------------------------
-- Copyright (c) 2016 CERN (BE-BI-QP)
-------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
-------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author
-- 2016-08-24  1.0      Jan Pospisil
------------------------------------------------------------------------------- 

library ieee;
use ieee.std_logic_1164.all;

entity Ad9512Syncer is
    generic (
        g_ClkFrequency: positive -- input clock frequency in Hz
    );
    port (
        Clk_ik: in std_logic;
        StartSync_i: in std_logic; -- one pulse starts synchronization
        Trigger_i: in std_logic; -- falling edge of this signals determines synchronization point
        Ad9512Sync_o: out std_logic := '0' -- synchronization pulse for AD9512
    );
end entity;

architecture syn of Ad9512Syncer is

    signal Pulse, GeneratorReset_ra: std_logic;

begin

    -- AD9512 function pin is set as SYNCB - needs negative pulse > 1.5*T_rfClk
    -- (f_rfClk >= 1 MHz) => (1.5*T_rfClk <= 1.5 us)
    
    cAd9512SyncPulseGenerator: entity work.PulseGeneratorTime(syn)
        generic map (
            g_ClkFrequency => g_ClkFrequency,
            g_PulseMinWidthInTime => 2 us
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => '0',
            Signal_i => StartSync_i,
            Pulse_o => Pulse
        );
    
    GeneratorReset_ra <= Trigger_i and (not Pulse);
    
    pSyncPulseGenerator: process(Clk_ik, GeneratorReset_ra) is begin
        if GeneratorReset_ra = '1' then
            Ad9512Sync_o <= '0';
        elsif rising_edge(Clk_ik) then
            if StartSync_i = '1' then
                Ad9512Sync_o <= '1';
            end if;
        end if;
    end process;

end architecture;