-------------------------------------------------------------------------------
-- Title      : Delayed Pulse Generator core
-- Project    : FMC DEL 1ns 2cha (FFPG)
-- URL        : http://www.ohwr.org/projects/fmc-del-1ns-2cha
-------------------------------------------------------------------------------
-- File       : DelayedPulseGenerator.vhd
-- Author(s)  : Jan Pospisil <j.pospisil@cern.ch>
-- Company    : CERN (BE-BI-QP)
-- Created    : 2016-07-05
-- Last update: 2016-08-24
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Main functionality of the pulse generation - read the memory
--              and outputs pulses, synchronized by Trigger_i.
-------------------------------------------------------------------------------
-- Copyright (c) 2016 CERN (BE-BI-QP)
-------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
-------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author        Comment
-- 2016-08-24  1.0      Jan Pospisil
-- 2016-09-07  1.1      Jan Pospisil  forced set/reset pulse registers to IOB
------------------------------------------------------------------------------- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.FfpgPkg.all;

entity DelayedPulseGenerator is
    port (
        -- main signals
        Clk_ik: in std_logic;
        Reset_ir: in std_logic;
        -- WB interface
        Overflow_ib16: in unsigned(15 downto 0);
        TriggerLatency_ib16: in unsigned(15 downto 0);
        Mode_i: in t_Mode;
        ModeLoad_i: in std_logic;
        SetMemAddress_ob11: out unsigned(10 downto 0);
        SetMemData_ib32: in unsigned(31 downto 0);
        SetMemReadStrobe_o: out std_logic;
        ResMemAddress_ob11: out unsigned(10 downto 0);
        ResMemData_ib32: in unsigned(31 downto 0);
        ResMemReadStrobe_o: out std_logic;
        Running_o: out std_logic;
        -- FMC interface
        Trigger_i: in std_logic;
        SetStream_o: out std_logic := '0';
        ResetStream_o: out std_logic := '0';
        -- debug
        FsmState_o: out unsigned(2 downto 0)
    );
end entity;

architecture syn of DelayedPulseGenerator is

    attribute iob: string;
    attribute iob of SetStream_o: signal is "FORCE";
    attribute iob of ResetStream_o: signal is "FORCE";

    constant c_MemLatency: integer := 2+1;
        -- +1 for memory output registers implemented here
    
    -- for memory output registers
    signal SetMemData_b32: unsigned(SetMemData_ib32'range) := (others => '0');
    signal ResMemData_b32: unsigned(ResMemData_ib32'range) := (others => '0');

    signal LastStreamPosition: unsigned(Overflow_ib16'range) := (others => '0');
    signal TriggerLatencyPlusTwo: unsigned(TriggerLatency_ib16'range) := (others => '0');

    signal GenerationEnable: std_logic;
    signal OutputEnable: std_logic;
    
    signal AddressEnableCounterOverflow: std_logic;
    signal AddressEnableCounterSetValue: unsigned(TriggerLatency_ib16'range);
    
    signal AddressCounterReset: std_logic := '0';
    signal AddressCounterResetOrReset: std_logic;
    signal AddressCounterSetValue: unsigned(TriggerLatency_ib16'range);
    signal AddressCounterValue: unsigned(SetMemAddress_ob11'range);
    
    signal StreamReset: std_logic := '0';
    signal StreamResetOrReset: std_logic;
    signal StreamPosition: unsigned(Overflow_ib16'range);

    signal BitCounterOverflow, LoadShiftRegister: std_logic;
    
    signal SetStream, ResetStream: std_logic;
    
begin

    -- memory output registers
    pMemOutputReg: process(Clk_ik) is begin
        if rising_edge(Clk_ik) then
            SetMemData_b32 <= SetMemData_ib32;
            ResMemData_b32 <= ResMemData_ib32;
        end if;
    end process;

    -- for better timing
    pInputReg: process(Clk_ik) is begin
        if rising_edge(Clk_ik) then
            LastStreamPosition <= Overflow_ib16 - 1;
            TriggerLatencyPlusTwo <= TriggerLatency_ib16 + 2;
                -- +1 because of ???
                -- +1 because of output registers on Set/ResetStream_o
        end if;
    end process;

    cFsm: entity work.Fsm(syn)
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Mode_i => Mode_i,
            ModeLoad_i => ModeLoad_i,
            Trigger_i => Trigger_i,
            StreamReset_i => StreamReset,
            Running_o => Running_o,
            GenerationEnable_o => GenerationEnable,
            OutputEnable_o => OutputEnable,
            State_o => FsmState_o
        );

    -- for better timing
    pAddressCounterReset: process(Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if StreamPosition = (LastStreamPosition - c_MemLatency) - 1 then
                AddressCounterReset <= '1';
            else
                AddressCounterReset <= '0';
            end if;
        end if;
    end process;
    
    AddressCounterResetOrReset <= AddressCounterReset or Reset_ir;
    AddressEnableCounterSetValue <= TriggerLatencyPlusTwo + c_MemLatency;
    cAddressEnableCounter: entity work.Counter(syn)
        generic map (
            g_Width => 5
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => AddressCounterResetOrReset,
            Enable_i => GenerationEnable,
            Set_i => Trigger_i,
            SetValue_ib => AddressEnableCounterSetValue(4 downto 0),
            Overflow_o => AddressEnableCounterOverflow,
            Value_ob => open
        );

    AddressCounterSetValue <= (TriggerLatencyPlusTwo + c_MemLatency);
    cAddressCounter: entity work.Counter(syn)
        generic map (
            g_Width => SetMemAddress_ob11'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => AddressCounterResetOrReset,
            Enable_i => AddressEnableCounterOverflow,
            Set_i => Trigger_i,
            SetValue_ib => AddressCounterSetValue(15 downto 5),
            Overflow_o => open,
            Value_ob => AddressCounterValue
        );
    
    SetMemAddress_ob11 <= AddressCounterValue;
    ResMemAddress_ob11 <= AddressCounterValue;
    
    SetMemReadStrobe_o <= '0'; -- this is not connected anywhere!!
    ResMemReadStrobe_o <= '0'; -- this is not connected anywhere!!
    
    -- for better timing
    pStreamResetReg: process(Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if StreamPosition = Overflow_ib16 - 2 then
                StreamReset <= '1';
            else
                StreamReset <= '0';
            end if;
        end if;
    end process;
    
    StreamResetOrReset <= StreamReset or Reset_ir;

    cStreamCounter: entity work.Counter(syn)
        generic map (
            g_Width => Overflow_ib16'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => StreamResetOrReset,
            Enable_i => GenerationEnable,
            Set_i => Trigger_i,
            SetValue_ib => TriggerLatencyPlusTwo,
            Overflow_o => open,
            Value_ob => StreamPosition
        );
    
    cBitCounter: entity work.Counter(syn)
        generic map (
            g_Width => 5
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => StreamResetOrReset,
            Enable_i => GenerationEnable,
            Set_i => Trigger_i,
            SetValue_ib => TriggerLatencyPlusTwo(4 downto 0),
            Overflow_o => BitCounterOverflow,
            Value_ob => open
        );
    
    LoadShiftRegister <= BitCounterOverflow or StreamReset;
    
    cShiftRegisterSet: entity work.ShiftRegister(right)
        generic map (
            g_Width => SetMemData_b32'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Data_ib => SetMemData_b32,
            Data_ob => open,
            Shift_i => '0',
            Shift_o => SetStream,
            ShiftEnable_i => GenerationEnable,
            LoadEnable_i => LoadShiftRegister
        );
    
    cShiftRegisterReset: entity work.ShiftRegister(right)
        generic map (
            g_Width => ResMemData_b32'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Data_ib => ResMemData_b32,
            Data_ob => open,
            Shift_i => '0',
            Shift_o => ResetStream,
            ShiftEnable_i => GenerationEnable,
            LoadEnable_i => LoadShiftRegister
        );
    
    pOutputRegs: process (Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if OutputEnable = '1' then
                SetStream_o <= SetStream;
                ResetStream_o <= ResetStream;
            else
                SetStream_o <= '0';
                ResetStream_o <= '0';
            end if;
        end if;
    end process;
    
end architecture;