-------------------------------------------------------------------------------
-- Title      : Clock domain crossing for Delayed Pulse Generator core
-- Project    : FMC DEL 1ns 2cha (FFPG)
-- URL        : http://www.ohwr.org/projects/fmc-del-1ns-2cha
-------------------------------------------------------------------------------
-- File       : DelayedPulseGeneratorsCdc.vhd
-- Author(s)  : Jan Pospisil <j.pospisil@cern.ch>
-- Company    : CERN (BE-BI-QP)
-- Created    : 2016-08-15
-- Last update: 2016-08-24
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Clock domain crossing between system (VME) clock and ClkRf_ik
--              RF clock.
-------------------------------------------------------------------------------
-- Copyright (c) 2016 CERN (BE-BI-QP)
-------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
-------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author
-- 2016-08-24  1.0      Jan Pospisil
-- 2016-09-05  1.1      Jan Pospisil  Independent Trigger Latency for the two
--                                    channels (issue 1389)
------------------------------------------------------------------------------- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.FfpgPkg.all;

entity DelayedPulseGeneratorsCdc is
    port (
        -- main signals
        Clk_ik: in std_logic;
        Reset_ir: in std_logic;
        -- trigger
        Trigger_i: in std_logic;
        -- configuration (WB)
        Ch1SetMemAddress_ob11: out unsigned(10 downto 0);
        Ch1SetMemData_ib32: in unsigned(31 downto 0);
        Ch1SetMemReadStrobe_o: out std_logic;
        Ch1ResMemAddress_ob11: out unsigned(10 downto 0);
        Ch1ResMemData_ib32: in unsigned(31 downto 0);
        Ch1ResMemReadStrobe_o: out std_logic;
        Ch2SetMemAddress_ob11: out unsigned(10 downto 0);
        Ch2SetMemData_ib32: in unsigned(31 downto 0);
        Ch2SetMemReadStrobe_o: out std_logic;
        Ch2ResMemAddress_ob11: out unsigned(10 downto 0);
        Ch2ResMemData_ib32: in unsigned(31 downto 0);
        Ch2ResMemReadStrobe_o: out std_logic;
        Overflow_ib16: in unsigned(15 downto 0);
        OverflowLoad_i: in std_logic;
        Ch1TriggerLatency_ib16: in unsigned(15 downto 0);
        Ch1TriggerLatencyLoad_i: in std_logic;
        Ch2TriggerLatency_ib16: in unsigned(15 downto 0);
        Ch2TriggerLatencyLoad_i: in std_logic;
        Ch1Mode_i: in t_Mode;
        Ch1Running_o: out std_logic;
        Ch2Mode_i: in t_Mode;
        Ch2Running_o: out std_logic;
        -- generation (FMC), synchronous to ClkRf_ik
        ClkRf_ik: in std_logic;
        Ch1SetStream_o: out std_logic;
        Ch1ResetStream_o: out std_logic;
        Ch2SetStream_o: out std_logic;
        Ch2ResetStream_o: out std_logic;
        -- debug
        Ch1FsmState_o: out unsigned(2 downto 0);
        Ch2FsmState_o: out unsigned(2 downto 0)
    );
end entity;

architecture syn of DelayedPulseGeneratorsCdc is

    signal Overflow_b_slv, OverflowRf_b_slv: std_logic_vector(Overflow_ib16'range) := (others => '0');
    signal Ch1TriggerLatency_b_slv, Ch1TriggerLatencyRf_b_slv: std_logic_vector(Ch1TriggerLatency_ib16'range) := (others => '0');
    signal CH2TriggerLatency_b_slv, Ch2TriggerLatencyRf_b_slv: std_logic_vector(Ch2TriggerLatency_ib16'range) := (others => '0');
    signal Ch1Mode_slv, Ch1ModeRf_slv: std_logic_vector(1 downto 0);
    signal Ch1ModeLoad: std_logic;
    signal Ch2Mode_slv, Ch2ModeRf_slv: std_logic_vector(1 downto 0);
    signal Ch2ModeLoad: std_logic;

    -- ClkRf clock domain
    signal ResetRf_r: std_logic;
    signal TriggerRf, TriggerRfPulse: std_logic;
    signal Ch1RunningRf: std_logic;
    signal Ch2RunningRf: std_logic;
    signal OverflowRf_b: unsigned(Overflow_ib16'range);
    signal Ch1TriggerLatencyRf_b: unsigned(Ch1TriggerLatency_ib16'range);
    signal Ch2TriggerLatencyRf_b: unsigned(Ch2TriggerLatency_ib16'range);
    signal Ch1ModeRf: t_Mode;
    signal Ch1ModeLoadRf: std_logic;
    signal Ch2ModeRf: t_Mode;
    signal Ch2ModeLoadRf: std_logic;

begin

    ----------------------------------
    -- CDC syncers to the ClkRf_ik clock domain
    ----------------------------------

    cResetSyncer: entity work.ResetSyncer(syn)
        port map (
            Clk_ik => ClkRf_ik,
            Reset_ira => Reset_ir,
            Reset_or => ResetRf_r
        );

    cTriggerSyncer: entity work.Delay(syn)
        generic map (
            g_Width => 1,
            g_Delay => 3,
            g_AsyncRegUsed => "TRUE"
        )
        port map (
            Clk_ik => ClkRf_ik,
            Data_ib(0) => Trigger_i,
            Data_ob(0) => TriggerRf
        );
    
    cTriggerEdgeDetector: entity work.EdgeDetector(rising)
        port map (
            Clk_ik => ClkRf_ik,
            Signal_i => TriggerRf,
            Edge_o => TriggerRfPulse
        );
    
    -- Overflow
    
    Overflow_b_slv <= std_logic_vector(Overflow_ib16);
    
    cOverflowSyncer: entity work.RegSyncer(syn)
        port map (
            ClkIn_ik => Clk_ik,
            ClkOut_ik => ClkRf_ik,
            Data_ib => Overflow_b_slv,
            Load_i => OverflowLoad_i,
            Data_ob => OverflowRf_b_slv,
            Load_o => open
        );
    
    OverflowRf_b <= unsigned(OverflowRf_b_slv);

    -- Ch1TriggerLatency
    
    Ch1TriggerLatency_b_slv <= std_logic_vector(Ch1TriggerLatency_ib16);
    
    cCh1TriggerLatencySyncer: entity work.RegSyncer(syn)
        port map (
            ClkIn_ik => Clk_ik,
            ClkOut_ik => ClkRf_ik,
            Data_ib => Ch1TriggerLatency_b_slv,
            Load_i => Ch1TriggerLatencyLoad_i,
            Data_ob => Ch1TriggerLatencyRf_b_slv,
            Load_o => open
        );
    
    Ch1TriggerLatencyRf_b <= unsigned(Ch1TriggerLatencyRf_b_slv);

    -- Ch2TriggerLatency
    
    Ch2TriggerLatency_b_slv <= std_logic_vector(Ch2TriggerLatency_ib16);
    
    cCh2TriggerLatencySyncer: entity work.RegSyncer(syn)
        port map (
            ClkIn_ik => Clk_ik,
            ClkOut_ik => ClkRf_ik,
            Data_ib => Ch2TriggerLatency_b_slv,
            Load_i => Ch2TriggerLatencyLoad_i,
            Data_ob => Ch2TriggerLatencyRf_b_slv,
            Load_o => open
        );
    
    Ch2TriggerLatencyRf_b <= unsigned(Ch2TriggerLatencyRf_b_slv);

    -- Ch1Mode
    
    Ch1Mode_slv <= f_ModeToSlv(Ch1Mode_i);
    
    cCh1ModeChange: entity work.ChangeDetector(syn)
        port map (
            Clk_ik => Clk_ik,
            Signal_ib => Ch1Mode_slv,
            Change_o => Ch1ModeLoad
        );
    
    cCh1ModeSyncer: entity work.RegSyncer(syn)
        port map (
            ClkIn_ik => Clk_ik,
            ClkOut_ik => ClkRf_ik,
            Data_ib => Ch1Mode_slv,
            Load_i => Ch1ModeLoad,
            Data_ob => Ch1ModeRf_slv,
            Load_o => Ch1ModeLoadRf
        );
    
    Ch1ModeRf <= f_SlvToMode(Ch1ModeRf_slv);

    -- Ch2Mode
    
    Ch2Mode_slv <= f_ModeToSlv(Ch2Mode_i);
    
    cCh2ModeChange: entity work.ChangeDetector(syn)
        port map (
            Clk_ik => Clk_ik,
            Signal_ib => Ch2Mode_slv,
            Change_o => Ch2ModeLoad
        );
    
    cCh2ModeSyncer: entity work.RegSyncer(syn)
        port map (
            ClkIn_ik => Clk_ik,
            ClkOut_ik => ClkRf_ik,
            Data_ib => Ch2Mode_slv,
            Load_i => Ch2ModeLoad,
            Data_ob => Ch2ModeRf_slv,
            Load_o => Ch2ModeLoadRf
        );
    
    Ch2ModeRf <= f_SlvToMode(Ch2ModeRf_slv);

    ----------------------------------
    -- main core in ClkRf_ik clock domain
    ----------------------------------

    cDelayedPulseGeneratorCh1: entity work.DelayedPulseGenerator(syn)
        port map (
            Clk_ik => ClkRf_ik,
            Reset_ir => ResetRf_r,
            Overflow_ib16 => OverflowRf_b,
            TriggerLatency_ib16 => Ch1TriggerLatencyRf_b,
            Mode_i => Ch1ModeRf,
            ModeLoad_i => Ch1ModeLoadRf,
            SetMemAddress_ob11 => Ch1SetMemAddress_ob11,
            SetMemData_ib32 => Ch1SetMemData_ib32,
            SetMemReadStrobe_o => Ch1SetMemReadStrobe_o,
            ResMemAddress_ob11 => Ch1ResMemAddress_ob11,
            ResMemData_ib32 => Ch1ResMemData_ib32,
            ResMemReadStrobe_o => Ch1ResMemReadStrobe_o,
            Running_o => Ch1RunningRf,
            Trigger_i => TriggerRfPulse,
            SetStream_o => Ch1SetStream_o,
            ResetStream_o => Ch1ResetStream_o,
            FsmState_o => Ch1FsmState_o
        );

    cDelayedPulseGeneratorCh2: entity work.DelayedPulseGenerator(syn)
        port map (
            Clk_ik => ClkRf_ik,
            Reset_ir => ResetRf_r,
            Overflow_ib16 => OverflowRf_b,
            TriggerLatency_ib16 => Ch2TriggerLatencyRf_b,
            Mode_i => Ch2ModeRf,
            ModeLoad_i => Ch2ModeLoadRf,
            SetMemAddress_ob11 => Ch2SetMemAddress_ob11,
            SetMemData_ib32 => Ch2SetMemData_ib32,
            SetMemReadStrobe_o => Ch2SetMemReadStrobe_o,
            ResMemAddress_ob11 => Ch2ResMemAddress_ob11,
            ResMemData_ib32 => Ch2ResMemData_ib32,
            ResMemReadStrobe_o => Ch2ResMemReadStrobe_o,
            Running_o => Ch2RunningRf,
            Trigger_i => TriggerRfPulse,
            SetStream_o => Ch2SetStream_o,
            ResetStream_o => Ch2ResetStream_o,
            FsmState_o => Ch2FsmState_o
        );

    ----------------------------------
    -- CDC syncer from the ClkRf_ik clock domain
    ----------------------------------

    cCh1RunningSyncer: entity work.Delay(syn)
        generic map (
            g_Width => 1,
            g_Delay => 3,
            g_AsyncRegUsed => "TRUE"
        )
        port map (
            Clk_ik => Clk_ik,
            Data_ib(0) => Ch1RunningRf,
            Data_ob(0) => Ch1Running_o
        );

    cCh2RunningSyncer: entity work.Delay(syn)
        generic map (
            g_Width => 1,
            g_Delay => 3,
            g_AsyncRegUsed => "TRUE"
        )
        port map (
            Clk_ik => Clk_ik,
            Data_ib(0) => Ch2RunningRf,
            Data_ob(0) => Ch2Running_o
        );

end architecture;