-------------------------------------------------------------------------------
-- Title      : FFPG core
-- Project    : FMC DEL 1ns 2cha (FFPG)
-- URL        : http://www.ohwr.org/projects/fmc-del-1ns-2cha
-------------------------------------------------------------------------------
-- File       : FfpgCore.vhd
-- Author(s)  : Jan Pospisil <j.pospisil@cern.ch>
-- Company    : CERN (BE-BI-QP)
-- Created    : 2016-06-28
-- Last update: 2016-08-24
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Core entity of the FFPG functionality ~ one FMC mezzanine.
--              Contains FFPG WB slave and OneWire WB slave for on-board
--              temperature sensor read-out.
-------------------------------------------------------------------------------
-- Copyright (c) 2016 CERN (BE-BI-QP)
-------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
-------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author
-- 2016-08-24  1.0      Jan Pospisil
------------------------------------------------------------------------------- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.FfpgPkg.all;
use work.ffpg_wbgen2_pkg.all;
use work.wishbone_pkg.all;

entity FfpgCore is
    generic (
        g_ClkFrequency: positive -- input clock frequency in Hz
    );
    port (
        -- Wishbone connection
        Clk_ik: in std_logic;
        Reset_ira: in std_logic; -- synchronized locally for better timing
        Wb_i: in t_wishbone_slave_in;
        Wb_o: out t_wishbone_slave_out;
        --- FMC interface
        -- clock
        ClkIn0_ik: in std_logic; -- 1 MHz <= f_ClkIn0_ik <= 204 MHz
        -- DACs
        TriggerDac_o: out t_Ad5600Interface;
        VcxoDac_o: out t_Ad5600Interface;
        -- output enable
        Ch1OutputEnable_o: out std_logic;
        Ch2OutputEnable_o: out std_logic;
        -- delay configuration
        DelayValue_ob: out unsigned(9 downto 0);
        Ch1SetLe_o: out std_logic;
        Ch1ResLe_o: out std_logic;
        Ch2SetLe_o: out std_logic;
        Ch2ResLe_o: out std_logic;
        --
        Ch1Set_o: out std_logic;
        Ch1Res_o: out std_logic;
        Ch2Set_o: out std_logic;
        Ch2Res_o: out std_logic;
        Trigger_i: in std_logic;
        -- AD9512 SPI
        SpiAd9512Sclk_o: out std_logic;
        SpiAd9512Mosi_o: out std_logic;
        SpiAd9512Miso_i: in std_logic;
        SpiAd9512Cs_on: out std_logic;
        -- AD9512 func. pin
        Ad9512Func_o: out std_logic;
        -- clock selection pin of SY58017U
        Clk2Sel_o: out std_logic;
        -- clock output
        ClkOut_ok: out std_logic;
        -- temperature chip interface
        Onewire_io: inout std_logic;
        -- LEDs
        Led_ob: out std_logic_vector(4 downto 1)
    );
end entity;

architecture syn of FfpgCore is

    -- Number of slave port(s) on the wishbone crossbar
    constant c_NumWbMasters: integer := 1;

    -- Number of master port(s) on the wishbone crossbar
    constant c_NumWbSlaves: integer := 3;

    -- Wishbone master(s)
    constant c_MasterId: integer := 0;

    -- Wishbone slave(s)
    constant c_SlaveTemperatureId: integer := 0;
    constant c_SlaveAd9512Id: integer := 1;
    constant c_SlaveFfpgId: integer := 2;

    -- sdb header address
    constant c_SdbAddress: t_wishbone_address := x"0000_0000";

    constant c_Ad9512Sdb : t_sdb_device := (
        abi_class     => x"0000",              -- undocumented device
        abi_ver_major => x"01",
        abi_ver_minor => x"01",
        wbd_endian    => c_sdb_endian_big,
        wbd_width     => x"7",                 -- 8/16/32-bit port granularity
        sdb_component => (
            addr_first  => x"0000000000000000",
            addr_last   => x"00000000000000ff",
            product     => (
                vendor_id => x"000000000000CE42",  -- CERN
                device_id => x"01010101",
                version   => x"00000001",
                date      => x"20170330",
                name      => "AD9512 SPI Ctrl.   "
            )
        )
    );

    -- Wishbone crossbar layout
    constant c_InterconnectLayout: t_sdb_record_array(c_NumWbSlaves-1 downto 0) := (
        0 => f_sdb_embed_device(c_xwb_onewire_master_sdb, x"0000_1000"),
        1 => f_sdb_embed_device(c_Ad9512Sdb,              x"0000_2000"),
        2 => f_sdb_embed_device(c_FfpgSdbDevice,          x"0001_0000")
    );

    signal Reset_r, Reset_nr: std_logic;

    -- Wishbone buse(s) from crossbar master port(s)
    signal CnxMasterOut: t_wishbone_master_out_array(c_NumWbSlaves-1 downto 0);
    signal CnxMasterIn: t_wishbone_master_in_array(c_NumWbSlaves-1 downto 0);

    -- Wishbone buse(s) to crossbar slave port(s)
    signal CnxSlaveOut: t_wishbone_slave_out_array(c_NumWbMasters-1 downto 0);
    signal CnxSlaveIn: t_wishbone_slave_in_array(c_NumWbMasters-1 downto 0);
    
    -- Onewire
    signal OwrEn, Owr: std_logic;
    
    -- SPI
    signal ChipSelect_b32: std_logic_vector(31 downto 0);
    
begin

    cResetSyncerSys: entity work.ResetSyncer(syn)
        port map (
            Clk_ik => Clk_ik,
            Reset_ira => Reset_ira,
            Reset_or => Reset_r
        );

    Reset_nr <= not Reset_r;

    ----------------------------------
    -- WB crossbar
    ----------------------------------

    cWbSdbCrossbar: entity work.xwb_sdb_crossbar(rtl)
        generic map (
            g_num_masters => c_NumWbMasters,
            g_num_slaves => c_NumWbSlaves,
            g_registered => true,
            g_wraparound => true,
            g_layout => c_InterconnectLayout,
            g_sdb_addr => c_SdbAddress
        )
        port map (
            clk_sys_i => Clk_ik,
            rst_n_i => Reset_nr,
            slave_i => CnxSlaveIn,
            slave_o => CnxSlaveOut,
            master_i => CnxMasterIn,
            master_o => CnxMasterOut
        );

    ----------------------------------
    -- WB Master connection
    ----------------------------------
    CnxSlaveIn(c_MasterId) <= Wb_i;
    Wb_o <= CnxSlaveOut(c_MasterId);

    ----------------------------------
    -- FFPG slave
    ----------------------------------
    
    cFfpgSlave: entity work.FfpgSlave(syn)
        generic map (
            g_ClkFrequency => g_ClkFrequency
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_r,
            Wb_i => CnxMasterOut(c_SlaveFfpgId),
            Wb_o => CnxMasterIn(c_SlaveFfpgId),
            ClkIn0_ik => ClkIn0_ik,
            TriggerDac_o => TriggerDac_o,
            VcxoDac_o => VcxoDac_o,
            Ch1OutputEnable_o => Ch1OutputEnable_o,
            Ch2OutputEnable_o => Ch2OutputEnable_o,
            DelayValue_ob => DelayValue_ob,
            Ch1SetLe_o => Ch1SetLe_o,
            Ch1ResLe_o => Ch1ResLe_o,
            Ch2SetLe_o => Ch2SetLe_o,
            Ch2ResLe_o => Ch2ResLe_o,
            Ch1Set_o => Ch1Set_o,
            Ch1Res_o => Ch1Res_o,
            Ch2Set_o => Ch2Set_o,
            Ch2Res_o => Ch2Res_o,
            Trigger_i => Trigger_i,
            Clk2Sel_o => Clk2Sel_o,
            Led_ob => Led_ob,
            SpiAd9512Cs_on => open,
            SpiAd9512Sclk_o => open,
            SpiAd9512Mosi_o => open,
            SpiAd9512Miso_i => '1',
            Ad9512Func_o => Ad9512Func_o,
            ClkOut_ok => ClkOut_ok
        );
    
    ----------------------------------
    -- 1-wire slave for temperature chip DS18B20
    ----------------------------------

    cSlaveTemperature: entity work.xwb_onewire_master(rtl)
        generic map (
            g_interface_mode => CLASSIC,
            g_address_granularity => BYTE,
            g_num_ports => 1,
            g_ow_btp_normal => "5.0",
            g_ow_btp_overdrive => "1.0",
            -- CDR_N = f_CLK * BTP_N - 1
            --   CDR_N = 125MHz * 5.0us - 1 = 625 - 1 = 624 = 0x0270
            -- CDR_O = f_CLK * BTP_O - 1
            --   CDR_O = 125MHz * 1.0us - 1 = 125 - 1 = 124 = 0x007C
            g_CDR_N => (g_ClkFrequency/1e6)*5 - 1,
            g_CDR_O => (g_ClkFrequency/1e6)*1 - 1
        )
        port map (
            clk_sys_i => Clk_ik,
            rst_n_i => Reset_nr,
            slave_i => CnxMasterOut(c_SlaveTemperatureId),
            slave_o => CnxMasterIn(c_SlaveTemperatureId),
            desc_o => open,
            owr_pwren_o => open,
            owr_en_o(0) => OwrEn,
            owr_i(0) => Owr
        );

    Onewire_io <=
        '0' when OwrEn = '1' else
        'Z';
    Owr <= Onewire_io;
        
    ----------------------------------
    -- AD9512 SPI slave
    ----------------------------------

    cSpiMasterWb: entity work.SpiMasterWb
    port map (
        Clk_ik => Clk_ik,
        Rst_irq => Reset_r,
        Cyc_i => CnxMasterOut(c_SlaveAd9512Id).cyc,
        Stb_i => CnxMasterOut(c_SlaveAd9512Id).stb,
        We_i => CnxMasterOut(c_SlaveAd9512Id).we,
        Adr_ib3 => CnxMasterOut(c_SlaveAd9512Id).adr(4 downto 2),
        Dat_ib32 => CnxMasterOut(c_SlaveAd9512Id).dat,
        Dat_oab32 => CnxMasterIn(c_SlaveAd9512Id).dat,
        Ack_oa => CnxMasterIn(c_SlaveAd9512Id).ack,
        WaitingNewData_o => open,
        ModuleIdle_o => open,
        SClk_o => SpiAd9512Sclk_o,
        MoSi_o => SpiAd9512Mosi_o,
        MiSo_ib32(0) => SpiAd9512Miso_i,
        MiSo_ib32(31 downto 1) => "0000000000000000000000000000000",
        SS_onb32 => ChipSelect_b32
    );
    CnxMasterIn(c_SlaveAd9512Id).err <= '0';
    CnxMasterIn(c_SlaveAd9512Id).rty <= '0';
    CnxMasterIn(c_SlaveAd9512Id).stall <= '0';
    CnxMasterIn(c_SlaveAd9512Id).int <= '0';
    SpiAd9512Cs_on <= ChipSelect_b32(0);

end architecture;
