-------------------------------------------------------------------------------
-- Title      : Clock frequency sense
-- Project    : FMC DEL 1ns 2cha (FFPG)
-- URL        : http://www.ohwr.org/projects/fmc-del-1ns-2cha
-------------------------------------------------------------------------------
-- File       : FrequencySense.vhd
-- Author(s)  : Jan Pospisil <j.pospisil@cern.ch>
-- Company    : CERN (BE-BI-QP)
-- Created    : 2016-08-23
-- Last update: 2016-08-24
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Calculate input clock frequency and determines its stability.
-------------------------------------------------------------------------------
-- Copyright (c) 2016 CERN (BE-BI-QP)
-------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
-------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author
-- 2016-08-24  1.0      Jan Pospisil
------------------------------------------------------------------------------- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity FrequencySense is
    generic (
        g_ClkFrequency: positive; -- input clock frequency in Hz
        g_SenseFrequencyWidth: positive; -- to accomodate maximum frequency to sense
        g_SenseFrequencyStableCount: natural := 0; -- how many measurements should be the "same"
            -- 0 ~ function disabled, SenseFrequencyStable_o always '0'
        g_SenseFrequencyStableWidth: positive := 1 -- how many MSBits are compared to be the "same"
            -- 32-10 ~ kHz, 32-20 ~ MHz
    );
    port (
        Clk_ik: in std_logic;
        SenseClk_ik: in std_logic;
        SenseFrequency_ob: out unsigned(g_SenseFrequencyWidth-1 downto 0);
        SenseFrequencyStable_o: out std_logic -- clock stable and present (f > 0)
    );
end entity;

architecture syn of FrequencySense is

    -- SD ~ SenseClk_ik Domain

    signal Latch, LatchToggle: std_logic := '0';
    signal LatchToggleSD, LatchSD: std_logic;
    signal
        FrequencyCounterValueSD_b, FrequencyCounterValueSD_bd,
        FrequencyCounterValue_b, FrequencyCounterValue_bd:
            std_logic_vector(g_SenseFrequencyWidth-1 downto 0);
    signal Frequency_b: unsigned(g_SenseFrequencyWidth-1 downto 0);
    signal SenseFrequency_b: unsigned(g_SenseFrequencyWidth-1 downto 0) := (others => '0');
    
    type t_ClkStable is array (0 to g_SenseFrequencyStableCount-1) of unsigned(g_SenseFrequencyStableWidth-1 downto 0);
    signal ClkRfFrequencyHistory: t_ClkStable := (others => (others => '0'));

begin

    ----------------------------------
    -- Clk_ik domain
    ----------------------------------

    c1HzGenerator: entity work.CounterLength(syn)
        generic map (
            g_Length => g_ClkFrequency
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => '0',
            Enable_i => '1',
            Overflow_o => Latch
        );

    pLatchToggle: process (Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if Latch = '1' then
                LatchToggle <= not LatchToggle;
            end if;
        end if;
    end process;
    
    cFrequencyCounterReg: entity work.Reg(syn)
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => '0',
            Data_ib => FrequencyCounterValue_b,
            Enable_i => Latch,
            Data_ob => FrequencyCounterValue_bd
        );
    
    Frequency_b <= unsigned(FrequencyCounterValue_b) - unsigned(FrequencyCounterValue_bd);

    pFrequencyReg: process (Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if Latch = '1' then
                SenseFrequency_b <= Frequency_b;
            end if;
        end if;
    end process;
    
    SenseFrequency_ob <= SenseFrequency_b;
    
    -- clock stability
    
    gClockStabilityEnabled: if g_SenseFrequencyStableCount > 0 generate

        ClkRfFrequencyHistory(0) <= SenseFrequency_b(g_SenseFrequencyWidth-1 downto g_SenseFrequencyWidth-g_SenseFrequencyStableWidth);
    
        pFrequencyHistory: process (Clk_ik) is begin
            if rising_edge(Clk_ik) then
                if Latch = '1' then
                    for i in 1 to g_SenseFrequencyStableCount-1 loop
                        ClkRfFrequencyHistory(i) <= ClkRfFrequencyHistory(i-1);
                    end loop;
                end if;
            end if;
        end process;
        
        pFrequencyStable: process (ClkRfFrequencyHistory) is
            variable Result: std_logic;
        begin
            Result := '1';
            for i in 1 to g_SenseFrequencyStableCount-1 loop
                if ClkRfFrequencyHistory(i) /= ClkRfFrequencyHistory(0) then
                    Result := '0';
                end if;
            end loop;
            if to_integer(ClkRfFrequencyHistory(0)) = 0 then
                Result := '0';
            end if;
            SenseFrequencyStable_o <= Result;
        end process;
    
    end generate;
    
    gClockStabilityDisabled: if g_SenseFrequencyStableCount = 0 generate
        SenseFrequencyStable_o <= '0';
    end generate;

    ----------------------------------
    -- CDC
    ----------------------------------
    
    cLatchSyncer: entity work.Delay(syn)
        generic map (
            g_Width => 1,
            g_Delay => 3,
            g_AsyncRegUsed => "TRUE"
        )
        port map (
            Clk_ik => SenseClk_ik,
            Data_ib(0) => LatchToggle,
            Data_ob(0) => LatchToggleSD
        );
    
    cFrequencyCounterSyncer: entity work.Delay(syn)
        generic map (
            g_Width => g_SenseFrequencyWidth,
            g_Delay => 3,
            g_AsyncRegUsed => "TRUE"
        )
        port map (
            Clk_ik => Clk_ik,
            Data_ib => FrequencyCounterValueSD_bd,
            Data_ob => FrequencyCounterValue_b
        );
    
    ----------------------------------
    -- SenseClk_ik domain
    ----------------------------------
    
    cFrequencyCounterSD: entity work.Counter(syn)
        generic map (
            g_Width => g_SenseFrequencyWidth
        )
        port map (
            Clk_ik => SenseClk_ik,
            Reset_ir => '0',
            Enable_i => '1',
            Set_i => '0',
            SetValue_ib => (others => '0'),
            Overflow_o => open,
            std_logic_vector(Value_ob) => FrequencyCounterValueSD_b
        );
    
    cLatchEdgeDetectorSD: entity work.EdgeDetector(both)
        port map (
            Clk_ik => SenseClk_ik,
            Signal_i => LatchToggleSD,
            Edge_o => LatchSD
        );
    
    cFrequencyCounterRegSD: entity work.Reg(syn)
        port map (
            Clk_ik => SenseClk_ik,
            Reset_ir => '0',
            Data_ib => FrequencyCounterValueSD_b,
            Enable_i => LatchSD,
            Data_ob => FrequencyCounterValueSD_bd
        );

end architecture;