--
-- RegSyncer
-- Can be used for CDC for vectors with sparse load signals - i.e. where there are at least several
-- clock cycles between successive load signals (e.g. Wishbone registers)
--

library ieee;
use ieee.std_logic_1164.all;

entity RegSyncer is
    port (
        -- main signals
        ClkIn_ik: in std_logic;
        ClkOut_ik: in std_logic;
        -- inputs
        Data_ib: in std_logic_vector;
        Load_i: in std_logic;
        -- outputs
        Data_ob: out std_logic_vector;
        Load_o: out std_logic
    );
end entity;

architecture syn of RegSyncer is

    signal Data_qb, DataSynced_b: std_logic_vector(Data_ib'range);
    signal NewData, NewDataSynced, NewDataSyncedPulse: std_logic := '0';
    signal ResetNewDataFlag, ResetNewDataPulse: std_logic;

begin

    -- data path
    cRegIn: entity work.Reg(syn)
        port map (
            Clk_ik => ClkIn_ik,
            Reset_ir => '0',
            Data_ib => Data_ib,
            Enable_i => Load_i,
            Data_ob => Data_qb
        );

    cDataSyncer: entity work.Delay(syn)
        generic map (
            g_Width => Data_ib'length,
            g_Delay => 3,
            g_AsyncRegUsed => "TRUE"
        )
        port map (
            Clk_ik => ClkOut_ik,
            Data_ib => Data_qb,
            Data_ob => DataSynced_b
        );
    
    cRegOut: entity work.Reg(syn)
        port map (
            Clk_ik => ClkOut_ik,
            Reset_ir => '0',
            Data_ib => DataSynced_b,
            Enable_i => NewDataSyncedPulse,
            Data_ob => Data_ob
        );

    -- control path
    
    pRs: process (ClkIn_ik) is begin
        if rising_edge(ClkIn_ik) then
            if ResetNewDataPulse = '1' then
                NewData <= '0';
            elsif Load_i = '1' then
                -- This load can be missed for processing if arrive to early, i.e. before
                -- ResetNewDataPulse corresponding to the previous Load_i.
                NewData <= '1';
            end if;
        end if;
    end process;
    
    cLoadSyncer: entity work.Delay(syn)
        generic map (
            g_Width => 1,
            g_Delay => 4,
            g_AsyncRegUsed => "TRUE"
        )
        port map (
            Clk_ik => ClkOut_ik,
            Data_ib(0) => NewData,
            Data_ob(0) => NewDataSynced
        );

    cNewDataSyncedPulseGenerator: entity work.EdgeDetector(rising)
        port map (
            Clk_ik => ClkOut_ik,
            Signal_i => NewDataSynced,
            Edge_o => NewDataSyncedPulse
        );
    
    cLoadDelay: entity work.Delay(syn)
        generic map (
            g_Width => 1,
            g_Delay => 2
        )
        port map (
            Clk_ik => ClkOut_ik,
            Data_ib(0) => NewDataSyncedPulse,
            Data_ob(0) => Load_o
        );

    cResetNewDataSyncer: entity work.Delay(syn)
        generic map (
            g_Width => 1,
            g_Delay => 3,
            g_AsyncRegUsed => "TRUE"
        )
        port map (
            Clk_ik => ClkIn_ik,
            Data_ib(0) => NewDataSynced,
            Data_ob(0) => ResetNewDataFlag
        );

    cResetNewDataPulseGenerator: entity work.EdgeDetector(rising)
        port map (
            Clk_ik => ClkIn_ik,
            Signal_i => ResetNewDataFlag,
            Edge_o => ResetNewDataPulse
        );
    
end architecture;