library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.RNG.all;

entity RegSyncer_tb is
end entity;

architecture testbench of RegSyncer_tb is

    -- 2.prime periods
    constant c_SlowClockPeriod: time := 106 ns;
    constant c_QuickClockPeriod: time := 10 ns;
    
    constant c_Width: positive := 8;
    constant c_Loops: positive := 100; -- number of test for each clock configuration (S-Q, Q-S)
    
    -- DUT signals
    signal ClkIn_ik, ClkOut_ik: std_logic := '0';
    signal Data_ib, Data_ob: std_logic_vector(c_Width-1 downto 0) := (others => 'X');
    signal Load_i, Load_o: std_logic := '0';

    -- TB signals
    signal ClkSlow, ClkQuick: std_logic := '0';
    signal SlowToQuick: std_logic := '0';
    signal Result_b: std_logic_vector(Data_ob'range) := (others => 'X');
    signal ResultReset: std_logic := '0';
    
    procedure f_SlowTick(ticks: in natural) is begin
		wait for ticks * c_SlowClockPeriod;
	end procedure;

    procedure f_QuiTick(ticks: in natural) is begin
		wait for ticks * c_QuickClockPeriod;
	end procedure;

    procedure f_InputTick(ticks: in natural) is begin
        if SlowToQuick then
            f_SlowTick(ticks);
        else
            f_QuiTick(ticks);
        end if;
    end procedure;

    procedure f_OutTick(ticks: in natural) is begin
        if SlowToQuick then
            f_QuiTick(ticks);
        else
            f_SlowTick(ticks);
        end if;
    end procedure;

begin

    cDUT: entity work.RegSyncer(syn)
        port map (
            ClkIn_ik => ClkIn_ik,
            ClkOut_ik => ClkOut_ik,
            Data_ib => Data_ib,
            Load_i => Load_i,
            Data_ob => Data_ob,
            Load_o => Load_o
        );
    
    pSlowGenerator: process is begin
        wait for c_SlowClockPeriod/2;
        ClkSlow <= not ClkSlow;
    end process;

    pQuickGenerator: process is begin
        wait for c_QuickClockPeriod/2;
        ClkQuick <= not ClkQuick;
    end process;
    
    ClkIn_ik <=
        ClkSlow when SlowToQuick else
        ClkQuick;
    ClkOut_ik <=
        ClkQuick when SlowToQuick else
        ClkSlow;
    
    pResultReg: process (ClkOut_ik) is begin
        if rising_edge(ClkOut_ik) then
            if ResultReset then
                Result_b <= (others => 'X');
            elsif Load_o then
                Result_b <= Data_ob;
            end if;
        end if;
    end process;
    
    pTest: process is

        variable Data_b: std_logic_vector(Data_ib'range) := (others => '0');
        variable Unf: Uniform := InitUniform(7, 0.0, (2.0**c_Width)-1.0);

        procedure f_RandomizeData is begin
            GenRnd(Unf);
            Data_b := std_logic_vector(to_unsigned(integer(Unf.rnd), c_Width));
        end procedure;

        procedure f_WriteData is begin
            -- reset result register
            ResultReset <= '1';
            f_OutTick(1);
            ResultReset <= '0';
            f_OutTick(1);
            -- write data
            Data_ib <= Data_b;
            Load_i <= '1';
            f_InputTick(1);
            Data_ib <= (others => 'X');
            Load_i <= '0';
        end procedure;
        
        procedure f_CheckData is begin
            assert Result_b = Data_b
                report "Data mismatch!"
                severity failure;
        end procedure;
        
    begin
        for i in 0 to 1 loop
            for j in 1 to c_Loops loop
                f_RandomizeData;
                f_InputTick(10);
                f_WriteData;
                f_OutTick(10);
                f_CheckData;
            end loop;
            SlowToQuick <= not SlowToQuick;
            f_SlowTick(10);
        end loop;

        assert false report "NONE. End of simulation." severity failure;
		wait;
    end process;

end architecture;