--
-- SlowToggle - slowly toggling signal(s)
-- v.2
-- by Jan Pospisil, j.pospisil@cern.ch, 2015
--
-- based on code
--   LED display driver - a parametrized function
--   by P. Jovanovic, 7.12.2005
--
-- Used mainly for visual LED indication of signal activity
-- Usage:
--   - generics:
--     g_Width: number of signals (inputs/outputs)
--     g_Ticks: number of Clk_ik period in one period of output change
--            (f=50MHz, T=20ns, desired output period t=100ms =>
--             g_Ticks=t/T=100ms/20ns=5000000)
--   - ports:
--     'Clk_ik' - main clock
--     'Reset_ir' - synchronous Reset_ir
--     'Test_i' - when asserted, all outputs are toggling
--     'Signal_ib' - input for observed signals
--     'Signal_ob' - output
--           - when there were any transition on input signal in last time t (last
--             g_Ticks period of 'Clk_ik') the output is toggling
--           - when there weren't any transition, output is the same as input (thus
--             signalling actual state of input)
--


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity SlowToggle is
	generic (
		g_Width: in positive := 8;
		g_Ticks: in positive := 256 -- output pulse width
	);
	port (
		Clk_ik: in std_logic;
		Reset_ir: in std_logic;
		Test_i: in std_logic;
		Signal_ib: in std_logic_vector(g_Width-1 downto 0);
		Signal_ob: out std_logic_vector(g_Width-1 downto 0)
	);
end entity;

architecture behavioral of SlowToggle is

	-- http://stackoverflow.com/a/12751341/615627
	function f_log2(x: positive) return natural is
		variable i: natural;
	begin
		i := 0;
		while (2**i < x) loop
			i := i + 1;
		end loop;
		return i;
	end function;	

	signal Counter_b: unsigned(f_log2(g_Ticks)-1 downto 0) := (others => '0');
	signal SignalInDelayed_b: std_logic_vector(g_Width-1 downto 0) := (others => '0');
	signal Edges_b, TestVector_b: std_logic_vector(g_Width-1 downto 0) := (others => '0');
	signal EdgeReset: std_logic := '0';
	signal Signal_b: std_logic_vector(g_Width-1 downto 0) := (others => '0');
	signal ChangeOutput: std_logic := '0';
    signal BlinkMaster_b: std_logic_vector(g_Width-1 downto 0) := (others => '0'); -- always blinking master signal

begin

	EdgeReset <= ChangeOutput or Reset_ir;
	TestVector_b <= (others => Test_i);

	pDelay: process (Clk_ik) is begin
		if rising_edge(Clk_ik) then
			SignalInDelayed_b <= Signal_ib;
		end if;
	end process;

	pEdgeAccumulator: process (Clk_ik) is begin
		if rising_edge(Clk_ik) then
			if EdgeReset = '1' then
				Edges_b <= (others => '0');
			else
				Edges_b <= Edges_b or (Signal_ib xor SignalInDelayed_b); -- latch positive or negative Edges_b
			end if;
		end if;
	end process;

	pCounter: process (Clk_ik) is begin
		if rising_edge(Clk_ik) then
			if Reset_ir = '1' then
				Counter_b <= (others => '0');
				ChangeOutput <= '0';
			else
				Counter_b <= Counter_b + 1;
				ChangeOutput <= '0';
				if Counter_b = to_unsigned(g_Ticks-1, Counter_b'length) then
					ChangeOutput <= '1';
					Counter_b <= (others => '0');
				end if;
			end if;
		end if;
	end process;

	pSignalOut: process (Clk_ik) is begin
		if rising_edge(Clk_ik) then
			if Reset_ir = '1' then
				Signal_b <= (others => '0');
			elsif ChangeOutput = '1' then
                BlinkMaster_b <= not BlinkMaster_b;
				Signal_b <= 
                    -- toggle if edge was present or Test_i is asserted
	                ((Edges_b or TestVector_b) and BlinkMaster_b)
	                or
                    -- otherwise sample signal
					(not (Edges_b or TestVector_b) and Signal_ib);
			end if;
		end if;
	end process;

	Signal_ob <= Signal_b;

end architecture;