/**
 * UVM verification package for AD5620/AD5640/AD5660 DAC
 * http://www.analog.com/media/en/technical-documentation/data-sheets/AD5620_5640_5660.pdf
 *   Revision G, 8/2013
 * 2016_06_03, Jan Pospisil, j.pospisil@cern.ch
 */

`include "uvm_macros.svh"

package Ad5600Package;

    import uvm_pkg::*;
    
    let MathMax(a, b) = (a > b) ? a : b;
    let MathMin(a, b) = (a < b) ? a : b;

    typedef enum {AD5600_TYPE_5620, AD5600_TYPE_5640, AD5600_TYPE_5660} t_ChipType;
    typedef enum {AD5600_SUFFIX_1, AD5600_SUFFIX_2, AD5600_SUFFIX_3} t_ChipSuffix;
    
    class Ad5600Chip;
        
        int DataBits[3] = {12, 14, 16};
        int RegisterBits[3] = {16, 16, 24};
        int DataStartInShiftRegister[3] = {2, 0, 0};
        
        t_ChipType ChipType = AD5600_TYPE_5620;
        t_ChipSuffix ChipSuffix = AD5600_SUFFIX_1;
        real PowerSupplyVoltage = 0;
        // VoltageHigh:
        //   0: V_DD = 2.7 V to 3.6 V, maximum SCL frequency = 20 MHz
        //   1: V_DD = 3.6 V to 5.5 V, maximum SCL frequency = 30 MHz
        bit VoltageHigh = 0;
        // ResetValueIsMidscale:
        //   0: output value after power-on reset is zero, part-name suffix: -1, -2
        //   1: output value after power-on reset is mid-scale, part-name suffix: -3
        bit ResetValueIsMidscale = 0;
        real VoltageReference = 0;
        string Name = "";
        
        function new(t_ChipType ChipType, t_ChipSuffix ChipSuffix, real PowerSupplyVoltage, string Name = "(unknown)");
            this.ChipType = ChipType;
            this.ChipSuffix = ChipSuffix;
            this.PowerSupplyVoltage = PowerSupplyVoltage;
            this.Name = Name;
            this.VoltageHigh = (PowerSupplyVoltage >= 3.6);
            this.ResetValueIsMidscale = (ChipSuffix == AD5600_SUFFIX_3);
            this.VoltageReference = ((ChipSuffix == AD5600_SUFFIX_1) ? 1.25 : 2.5);
        endfunction
        
        function int getDataBits;
            return this.DataBits[this.ChipType];
        endfunction
        
        function int getRegisterBits;
            return this.RegisterBits[this.ChipType];
        endfunction
        
        function int getDataStartInShiftRegister;
            return this.DataStartInShiftRegister[this.ChipType];
        endfunction
        
        function real getOutputVoltage(int code);
            real voltage;
            code = code % (2 ** this.getDataBits()); // assert?
            voltage = 2 * this.VoltageReference * code / (2 ** this.getDataBits());
            voltage = MathMax(voltage, 0);
            voltage = MathMin(voltage, this.PowerSupplyVoltage);
            return voltage;
        endfunction
        
    endclass
    
    class Ad5600Config extends uvm_object;
        `uvm_object_utils(Ad5600Config)

        virtual Ad5600Interface VirtualInterface;
        Ad5600Chip Chip;

        function new(string name = "Ad5600Config");
            super.new(name);
        endfunction: new

    endclass
    
    class Ad5600Transaction extends uvm_sequence_item;
        rand int VoltageRegister;
        Ad5600Chip Chip;
        
        `uvm_object_utils(Ad5600Transaction)

        constraint VoltageRegisterConstrain {
            VoltageRegister >= 0;
            VoltageRegister < 2**Chip.getDataBits();
        }

        function new (string name = "");
            super.new(name);
        endfunction

        function string convert2string;
            return $sformatf("Output voltage on chip \"%s\" changed: REG = 0x%H (%0d), V_out = %f V", this.Chip.Name, this.VoltageRegister, this.VoltageRegister, this.Chip.getOutputVoltage(this.VoltageRegister));
        endfunction

    endclass
    
    class Ad5600Monitor extends uvm_monitor;
        `uvm_component_utils(Ad5600Monitor)

        uvm_analysis_port #(Ad5600Transaction) AnalysisPort;
        Ad5600Config Configuration;
        virtual Ad5600Interface.tb LocalInterface;
        Ad5600Chip Chip;
        
        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction
        
        function void build_phase(uvm_phase phase);
            AnalysisPort = new("AnalysisPort", this);
            assert(uvm_config_db #(Ad5600Config)::get(this, "", "Ad5600Config", Configuration)) else
                `uvm_fatal("CONF", "Configuration Ad5600Config not found!");
            LocalInterface = Configuration.VirtualInterface;
            Chip = Configuration.Chip;
        endfunction

        task run_phase(uvm_phase phase);
            int dataBits = this.Chip.getDataBits();
            int registerBits = this.Chip.getRegisterBits();
            bit [31:0] shiftRegister;
            Ad5600Transaction tx;
            int dataStart = this.Chip.getDataStartInShiftRegister();
            forever begin
                @(negedge LocalInterface.FrameSynchronization_n);
                phase.raise_objection(this);
                repeat (registerBits) begin
                    @(negedge LocalInterface.SerialClock);
                    // TODO: maximal clock speed assert
                    shiftRegister = {shiftRegister[0 +: 31], LocalInterface.SerialData};
                    if (LocalInterface.FrameSynchronization_n)
                        break;
                end
                if (~LocalInterface.FrameSynchronization_n) begin
                    tx = Ad5600Transaction::type_id::create("tx");
                    tx.VoltageRegister = int'((shiftRegister >> dataStart) & 2**(dataBits)-1);
                    // TODO: power modes?
                    tx.Chip = this.Chip;
                    AnalysisPort.write(tx);
                    @(posedge LocalInterface.FrameSynchronization_n);
                end
                phase.drop_objection(this);
            end
        endtask

    endclass
    
    class Ad5600Agent extends uvm_agent;
        `uvm_component_utils(Ad5600Agent)
        
        uvm_analysis_port #(Ad5600Transaction) AnalysisPort;
        Ad5600Monitor Ad5600Monitor_h;

        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction

        function void build_phase(uvm_phase phase);
            AnalysisPort = new("AnalysisPort", this);
            Ad5600Monitor_h = Ad5600Monitor::type_id::create("Ad5600Monitor_h", this);
        endfunction

        function void connect_phase(uvm_phase phase);
            Ad5600Monitor_h.AnalysisPort.connect(AnalysisPort);
        endfunction
        
    endclass
    
endpackage