//-----------------------------------------------------------------------------
// Title      : UVM verification package for FMC DEL 1ns 2cha (FFPG)
// Project    : FMC DEL 1ns 2cha (FFPG)
// URL        : http://www.ohwr.org/projects/fmc-del-1ns-2cha
//-----------------------------------------------------------------------------
// File       : FfpgFmcPackage.sv
// Author(s)  : Jan Pospisil <j.pospisil@cern.ch>
// Company    : CERN (BE-BI-QP)
// Created    : 2016-07-06
// Last update: 2016-08-24
// Standard   : SystemVerilog, UVM
//-----------------------------------------------------------------------------
// Description: 
//-----------------------------------------------------------------------------
// Copyright (c) 2016 CERN (BE-BI-QP)
//-----------------------------------------------------------------------------
// GNU LESSER GENERAL PUBLIC LICENSE
//-----------------------------------------------------------------------------
// This source file is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published by the
// Free Software Foundation; either version 2.1 of the License, or (at your
// option) any later version. This source is distributed in the hope that it
// will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details. You should have
// received a copy of the GNU Lesser General Public License along with this
// source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
//-----------------------------------------------------------------------------
// Revisions  :
// Date        Version  Author
// 2016-08-24  1.0      Jan Pospisil
//----------------------------------------------------------------------------- 

`include "timescale.svh"
`include "uvm_macros.svh"

package FfpgFmcPackage;

    import uvm_pkg::*;

    class FfpgFmcConfig extends uvm_object;
        `uvm_object_utils(FfpgFmcConfig)

        virtual FfpgFmcInterface VirtualInterface;
        int ClockPeriod; // in time units
        int TriggerPeriodicity; // in clock cycles
        int TriggerPhase; // in time units
        int TriggerLatency; // in clock cycles

        function new(string name = "FfpgFmcConfig");
            super.new(name);
        endfunction: new

    endclass
    
    class FfpgFmcDriverTransaction extends uvm_sequence_item;
        `uvm_object_utils(FfpgFmcDriverTransaction)

        rand int Cycles; // how many turn cycles to simulate (0 ~ forever)
        
        constraint CyclesConstraint {
            Cycles inside {[3:20]};
        }

        function new (string name = "");
            super.new(name);
        endfunction

        function string convert2string;
            return $sformatf("Running %0d turns.", this.Cycles);
        endfunction

    endclass
    
    class FfpgFmcPulseTransaction extends uvm_sequence_item;
        `uvm_object_utils(FfpgFmcPulseTransaction)
        
        int BunchCount;
        rand int Channel;
        rand int StartBunch;
        rand int StopBunch;
        
        constraint ChannelConstraint {
            Channel inside {[1:2]};
        }
        
        constraint StartBunchConstrain {
            StartBunch >= 0;
            StartBunch < BunchCount;
        }
        
        constraint StopBunchConstrain {
            StopBunch >= StartBunch;
            StopBunch < BunchCount;
        }
        
        function new(string name = "");
            super.new(name);
        endfunction

        function string convert2string;
            if (this.StartBunch == this.StopBunch)
                return $sformatf("Channel %0d: pulse @ bunch %0d.", this.Channel, this.StartBunch);
            else
                return $sformatf("Channel %0d: pulse spanning from bunch %0d to bunch %0d.", this.Channel, this.StartBunch, this.StopBunch);
        endfunction
        
        virtual function bit do_compare(uvm_object rhs, uvm_comparer comparer);
            FfpgFmcPulseTransaction rhs_;
            $cast(rhs_,rhs);
            do_compare = 1;
            do_compare &= (this.Channel == rhs_.Channel);
            do_compare &= (this.StartBunch == rhs_.StartBunch);
            do_compare &= (this.StopBunch == rhs_.StopBunch);
        endfunction

    endclass
    
    class FfpgFmcMonitor extends uvm_monitor;
        `uvm_component_utils(FfpgFmcMonitor)

        uvm_analysis_port #(FfpgFmcPulseTransaction) AnalysisPort;
        FfpgFmcConfig Configuration;
        virtual FfpgFmcInterface LocalInterface;
        
        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction
        
        function void build_phase(uvm_phase phase);
            AnalysisPort = new("AnalysisPort", this);
            assert(uvm_config_db #(FfpgFmcConfig)::get(this, "", "FfpgFmcConfig", Configuration)) else
                `uvm_fatal("CONF", "Configuration FfpgFmcConfig not found!");
            LocalInterface = Configuration.VirtualInterface;
        endfunction

        function int GetRealBunch(int Bunch);
            int Result = Bunch;
            
            // TriggerLatency
            Result = Result + Configuration.TriggerLatency;
            
            // Trigger_i synchronisation delay
            Result = Result - 3;
            
            return (Result + Configuration.TriggerPeriodicity) % Configuration.TriggerPeriodicity;
        endfunction
        
        task run_phase(uvm_phase phase);
            int Bunch = 0;
            bit PulsePresentCh1;
            bit PulsePresentCh2;
            logic Ch1SetOld = 0;
            logic Ch1ResOld = 0;
            logic Ch2SetOld = 0;
            logic Ch2ResOld = 0;
            FfpgFmcPulseTransaction TxCh1 = null;
            FfpgFmcPulseTransaction TxCh2 = null;

            forever begin
                @(posedge LocalInterface.ClkIn0_k);
                
                if (LocalInterface.Trigger)
                    Bunch = 0;

                // rising edge sensitive actions

                if (~PulsePresentCh1 && ~Ch1SetOld && LocalInterface.Ch1Set) begin
                    phase.raise_objection(this);
                    PulsePresentCh1 = 1;
                    TxCh1 = FfpgFmcPulseTransaction::type_id::create("TxCh1");
                    TxCh1.StartBunch = GetRealBunch(Bunch);
                    TxCh1.Channel = 1;
                end
                if (PulsePresentCh1 && ~Ch1ResOld && LocalInterface.Ch1Res) begin
                    TxCh1.StopBunch = GetRealBunch(Bunch);
                    AnalysisPort.write(TxCh1);
                    PulsePresentCh1 = 0;
                    phase.drop_objection(this);
                end
                
                if (~PulsePresentCh2 && ~Ch2SetOld && LocalInterface.Ch2Set) begin
                    phase.raise_objection(this);
                    PulsePresentCh2 = 1;
                    TxCh2 = FfpgFmcPulseTransaction::type_id::create("TxCh2");
                    TxCh2.StartBunch = GetRealBunch(Bunch);
                    TxCh2.Channel = 2;
                end
                if (PulsePresentCh2 && ~Ch2ResOld && LocalInterface.Ch2Res) begin
                    TxCh2.StopBunch = GetRealBunch(Bunch);
                    AnalysisPort.write(TxCh2);
                    PulsePresentCh2 = 0;
                    phase.drop_objection(this);
                end
                
                Ch1SetOld = LocalInterface.Ch1Set;
                Ch1ResOld = LocalInterface.Ch1Res;
                Ch2SetOld = LocalInterface.Ch2Set;
                Ch2ResOld = LocalInterface.Ch2Res;
                
                Bunch = Bunch + 1;
            end
        endtask

    endclass
   
    class FfpgFmcDriver extends uvm_driver #(FfpgFmcDriverTransaction);
        `uvm_component_utils(FfpgFmcDriver)
        
        FfpgFmcConfig Configuration;
        virtual FfpgFmcInterface LocalInterface;
        
        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction
        
        function void build_phase(uvm_phase phase);
            assert(uvm_config_db #(FfpgFmcConfig)::get(this, "", "FfpgFmcConfig", Configuration)) else
                `uvm_fatal("CONF", "Configuration FfpgFmcConfig not found!");
            LocalInterface = Configuration.VirtualInterface;
        endfunction

        task run_phase(uvm_phase phase);
            forever begin
                FfpgFmcDriverTransaction tx;
                int Cycle;
                int Bunch;
                
                seq_item_port.get_next_item(tx);
                `uvm_info("FMCGEN", tx.convert2string(), UVM_NONE);
                if (tx.Cycles > 0)
                    phase.raise_objection(this);
                if (Configuration.ClockPeriod % 2)
                    $warning("Clock period is not a multiple of 2, different clock period is generated!");
                LocalInterface.ClkIn0_k <= 0;
                LocalInterface.Trigger <= 0;
                
                for (Cycle = 0; (Cycle < tx.Cycles) || (tx.Cycles == 0); Cycle++ ) begin
                    for (Bunch = 0; Bunch < Configuration.TriggerPeriodicity; Bunch++) begin
                        fork
                            begin
                                // clock generation
                                LocalInterface.ClkIn0_k <= 1;
                                #(Configuration.ClockPeriod/2);
                                LocalInterface.ClkIn0_k <= 0;
                                #(Configuration.ClockPeriod/2);
                            end
                            begin
                                // trigger generation
                                if (Bunch == 0) begin
                                    #Configuration.TriggerPhase;
                                    LocalInterface.Trigger <= 1;
                                end
                                if (Bunch == 1) begin
                                    #Configuration.TriggerPhase;
                                    LocalInterface.Trigger <= 0;
                                end
                            end
                        join
                    end
                end
                if (tx.Cycles > 0)
                    phase.drop_objection(this);
                seq_item_port.item_done();
            end
        endtask

    endclass
    
    class FfpgFmcAgent extends uvm_agent;
        `uvm_component_utils(FfpgFmcAgent)
        
        uvm_analysis_port #(FfpgFmcPulseTransaction) AnalysisPort;
        FfpgFmcMonitor Monitor_h;
        FfpgFmcDriver Driver_h;
        uvm_sequencer #(FfpgFmcDriverTransaction) Sequencer_h;

        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction

        function void build_phase(uvm_phase phase);
            FfpgFmcConfig Configuration;
            
            assert(uvm_config_db #(FfpgFmcConfig)::get(this, "", "FfpgFmcConfig", Configuration)) else
                `uvm_fatal("CONF", "Configuration FfpgFmcConfig not found!");
            `uvm_info(
                "FMCAGENT",
                $sformatf(
                    "Building FMC agent: clock period: %0d ns (%.3f MHz), trigger periodicity: %0d cycles, trigger phase: %0d ns, trigger latency %0d cycles",
                    Configuration.ClockPeriod, 1000.0/Configuration.ClockPeriod, Configuration.TriggerPeriodicity, Configuration.TriggerPhase, Configuration.TriggerLatency),
                UVM_NONE)

            AnalysisPort = new("AnalysisPort", this);
            Monitor_h = FfpgFmcMonitor::type_id::create("Monitor_h", this);
            Driver_h = FfpgFmcDriver::type_id::create("Driver_h", this);
            Sequencer_h = uvm_sequencer #(FfpgFmcDriverTransaction)::type_id::create("Sequencer_h", this);
        endfunction

        function void connect_phase(uvm_phase phase);
            Monitor_h.AnalysisPort.connect(AnalysisPort);
            Driver_h.seq_item_port.connect(Sequencer_h.seq_item_export);
        endfunction
        
    endclass
    
endpackage