##-----------------------------------------------------------------------------
## Title      : Python drive: FFPG driver class
## Project    : FMC DEL 1ns 2cha (FFPG)
## URL        : http://www.ohwr.org/projects/fmc-del-1ns-2cha
##-----------------------------------------------------------------------------
## File       : Ffpg.py
## Author(s)  : Jan Pospisil <j.pospisil@cern.ch>
## Company    : CERN (BE-BI-QP)
## Created    : 2016-09-06
## Last update: 2016-09-06
## Standard   : Python
##-----------------------------------------------------------------------------
## Description:
##-----------------------------------------------------------------------------
## Copyright (c) 2016 CERN (BE-BI-QP)
##-----------------------------------------------------------------------------
## GNU LESSER GENERAL PUBLIC LICENSE
##-----------------------------------------------------------------------------
## This source file is free software; you can redistribute it and/or modify it
## under the terms of the GNU Lesser General Public License as published by the
## Free Software Foundation; either version 2.1 of the License, or (at your
## option) any later version. This source is distributed in the hope that it
## will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
## of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU Lesser General Public License for more details. You should have
## received a copy of the GNU Lesser General Public License along with this
## source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
##-----------------------------------------------------------------------------
## Revisions  :
## Date        Version  Author        Comment
## 2016-09-06  1.0      Jan Pospisil  Derived from the initial example driver
## 2016-09-07  1.1      Jan Pospisil  added AD9512 OUT4 fine delay
## 2016-09-21  1.2      Jan Pospisil  added GetFmcSlot and LED test methods
##-----------------------------------------------------------------------------

import time
import warnings
from WbOverlay import *
from DS18B20 import *
from AD9512 import *
from CommonFunctions import *

class Ffpg(object):

    # TODO: proper BUSY bits checking

    wb = None
    inputClockFrequency = None # frequency (in MHz) of the selected clock
    fineTriggerPhase = {1: 0, 2: 0} # individual for each channel, in [ns]
    temp = None # DS18B20 OneWire temperature chip
    ad9512 = None # AD9512 SPI
    fmcSlot = None
    
    def __init__(self, fmcSlot):
        """ fmcSlot = 0 | 1 """
        assert (fmcSlot&1) == fmcSlot, "fmcSlot can be just 0 or 1!"
        self.fmcSlot = fmcSlot
        self.wb = WbOverlay("fmc_fpg", fmcSlot)
        self.temp = DS18B20(wbOverlay = self.wb)
        self.ad9512 = AD9512(wbOverlay = self.wb, SpiBaseAddress = "SpiBase")
        assert self.CompareVersion(1, 3, 0) == 0, "Bad gateware version detected! Get the appropriate driver from http://www.ohwr.org/projects/fmc-del-1ns-2cha/repository/revisions/master/show/sw"

    def GetFmcSlot(self):
        return self.fmcSlot
        
    def GetVersion(self):
        version = self.wb.Read("version")
        major = (version >> 22) & ((2**10)-1)
        minor = (version >> 12) & ((2**10)-1)
        revision = version & ((2**12)-1)
        return [major, minor, revision]

    def PrintVersion(self):
        [major, minor, revision] = self.GetVersion()
        print("Gateware version: "+str(major)+"."+str(minor)+"."+str(revision))

    def CompareVersion(self, major, minor, revision):
        """
        check actual version against provided version
        returns:
           0 - versions are same
           1 - actual version is newer than argument provided
          -1 - actual version is older than argument provided
        """
        [actualMajor, actualMinor, actualRevision] = self.GetVersion()
        if actualMajor > major:
            return 1
        if actualMajor < major:
            return -1
        if actualMinor > minor:
            return 1
        if actualMinor < minor:
            return -1
        if actualRevision > revision:
            return 1
        if actualRevision < revision:
            return -1
        return 0

    def SelectClock(self, clockType, frequency):
        """
        clockType = "external" | "fpga" | "vcxo"
        frequency in [MHz]
        """
        bits = 0
        if clockType == "external":
            bits = 0
        elif clockType == "fpga":
            bits = 1
        elif clockType == "vcxo":
            bits = 2
        else:
            raise Exception("Unknown clock \""+clockType+"\"!")
        self.wb.SetBits("control", 3, bits)
        self.inputClockFrequency = frequency

    def SetRatio(self, ratio):
        """ ratio = 1-32 """
        if (ratio < 1) or (ratio > 32):
            raise Exception("Bad ratio!")
        self.wb.Write("clock_ratio_m1", ratio-1)

    def GetRatio(self):
        ratiom1 = self.wb.Read("clock_ratio_m1")
        return ratiom1+1
    
    def SetOverflow(self, overflow):
        self.wb.Write("overflow", overflow)

    def Ad9512Sync(self):
        """ synchronize AD9512 dividers """
        self.wb.SetBits("control", 1<<9)

    def SetTriggerThreshold(self, voltage):
        """ voltage in [V] """
        vdd = 5
        bits = 16
        code = int((voltage/float(vdd)) * (2**bits))
        code &= (2**bits)-1
        self.wb.Write("trig_threshold", code)

    def SetVcxoFrequency(self, compensation):
        """ compensation = <-1;1) """
        vdd = 3.3
        bits = 16
        voltage = vdd * (compensation + 1) / 2.0
        code = int((voltage/float(5)) * (2**bits))
        code &= (2**bits)-1
        self.wb.Write("vcxo_voltage", code)

    def GetFrequency(self):
        # return frequency of RF clock in Hz, if stable, None otherwise
        status = self.wb.Read("status")
        if (status & (1<<8) == 0):
            return None
        else:
            return self.wb.Read("frequency")

    def ClearMemory(self, channel, memory, memoryPart = -1):
        """
        channel = 1 | 2 | 0 (for both)
        memory = 'set' | 'res' | '*' (for both memories)
        if memoryPart > -1: clear only first memoryPart words
        """
        if channel == 0:
            self.ClearMemory(1, memory, memoryPart)
            self.ClearMemory(2, memory, memoryPart)
            return
        
        assert (memory == "set") or (memory == "res") or (memory == "*"), "Bad memory!"
        if memory == "*":
            self.ClearMemory(channel, "set", memoryPart)
            self.ClearMemory(channel, "res", memoryPart)
            return
        
        partToClear = 2048
        to = -1
        if memoryPart > -1:
            partToClear = memoryPart
            to = memoryPart
        self.wb.WriteMulti(self._GetChannelRegName(channel, memory+"_mem"), (0,)*partToClear, to = to)

    def Reset(self):
        self.ClearMemory(0, "*")
        self.wb.Write("control", 0)

    def _GetChannelRegName(self, channel, name):
        assert (channel == 1) or (channel == 2), "Bad channel \"" + str(channel) + "\", we have only 2 channels!"
        return "ch" + str(channel) + "_" + name

    def GetClockPeriod(self):
        """ return clock period in [ns] of the RF clock coming to the FPGA """
        return 1000.0/(self.inputClockFrequency/self.GetRatio())

    def SetFineDelay(self, channel, memory, delayValue):
        """
        channel = 1 | 2
        memory = "set" | "res"
        delayValue in [ns]
        """
        # compensate for calibration
        delayValue += self.fineTriggerPhase[channel]
        # calculate delay binary value
        binaryValue = int(round(delayValue * 100)) # step is 10 ps
        assert binaryValue < 2**10, "Fine delay is larger than the IC can handle!"
        # set delay
        self.wb.Write(self._GetChannelRegName(channel, "delay_"+memory), binaryValue)

    def SetTriggerPhase(self, channel, phase):
        """ phase in [ns] - depends on setup, cable length... """
        clockPeriod = self.GetClockPeriod()
        bitPhase = int(phase/clockPeriod)           # advance output pulse
        finePhase = phase - bitPhase * clockPeriod  # delay output pulse
        if finePhase > 0:
            bitPhase += 1
            finePhase = clockPeriod - finePhase
        self.wb.Write(self._GetChannelRegName(channel, "trig_latency"), bitPhase)
        self.fineTriggerPhase[channel] = finePhase
        self.SetFineDelay(channel, "set", 0)

    def PrintStatus(self):
        print("Status register:")
        status = self.wb.Read("status")
        PrintBit(status, 0, "Clock infrastructure configuration", "busy", "idle")
        PrintBit(status, 1, "VCXO DAC", "busy", "idle")
        PrintBit(status, 2, "Trigger threshold DAC", "busy", "idle")
        PrintBit(status, 3, "Delay configuration", "busy", "idle")
        PrintBit(status, 4, "Channel 1 output", "enabled", "disabled")
        PrintBit(status, 5, "Channel 2 output", "enabled", "disabled")
        PrintBit(status, 6, "Channel 1", "running", "stopped")
        PrintBit(status, 7, "Channel 2", "running", "stopped")
        PrintBit(status, 8, "Input clock stability", "stable", "not stable/present")

    def PrintControl(self):
        print("Control register: ")
        control = self.wb.Read("control")
        PrintBits(control, 0, 1, "Clock selection", ("external clock", "FPGA LOOP", "on-board VCXO"))
        PrintBit(control, 2, "Channel 1 output", "enabled", "disabled")
        PrintBit(control, 3, "Channel 2 output", "enabled", "disabled")
        PrintBits(control, 4, 5, "Channel 1 mode", ("stopped", "continuous", "one-shot"))
        PrintBits(control, 6, 7, "Channel 2 mode", ("stopped", "continuous", "one-shot"))
        PrintBit(control, 8, "LEDs", "blinking", "normal operation")
        PrintBit(control, 9, "AD9512 Synchronization", "in progress", "done")

    def PrintDebug(self):
        print("Debug:")
        debug = self.wb.Read("debug")
        PrintBits(debug, 0, 2, "CH1 FSM state", ("Stop", "WaitForTrigger", "Generating", "Outputting"))
        PrintBits(debug, 3, 5, "CH2 FSM state", ("Stop", "WaitForTrigger", "Generating", "Outputting"))

    def EnableChannel(self, channel):
        channel -= 1
        channel %= 2
        self.wb.SetBits("control", 0x4<<channel, 0x4<<channel) # set output enable

    def DisableChannel(self, channel):
        channel -= 1
        channel %= 2
        self.wb.SetBits("control", 0x4<<channel, 0) # set output disable

    def StartChannel(self, channel):
        channel -= 1
        channel %= 2
        channel *= 2
        self.wb.SetBits("control", 0x30<<channel, 0x10<<channel) # set mode CONT.

    def StartChannelOnce(self, channel):
        channel -= 1
        channel %= 2
        channel *= 2
        self.wb.SetBits("control", 0x30<<channel, 0x20<<channel) # set mode ONE-SHOT

    def StopChannel(self, channel):
        channel -= 1
        channel %= 2
        channel *= 2
        self.wb.SetBits("control", 0x30<<channel, 0) # set mode STOP

    def _ConfigurePulse(self, channel, setIndex, resIndex, pulse, which = 0):
        """
         Manipulate the SET and RESET pulse generator arrays:
            
         channel: 1, 2, 0 (for both)
         setIndex/resIndex: bit-wise memory address
         pulse: 0 - delete, 1 - create
         which: 0 - both, 1 - set, 2 - reset

                        |                which: 0                 |
                        |      which: 1     |      which: 2       |
            pulse: 0    | SET[setIndex] = 0 | RESET[resIndex] = 0 |
            pulse: 1    | SET[setIndex] = 1 | RESET[resIndex] = 1 |
            
        """
        if channel == 0:
            self._ConfigurePulse(1, setIndex, resIndex, pulse, which)
            self._ConfigurePulse(2, setIndex, resIndex, pulse, which)
            return
        
        if (setIndex < 0) or (setIndex > 65535):
            raise Exception("Bad setIndex!")
        if (resIndex < 0) or (resIndex > 65535):
            raise Exception("Bad resIndex!")
        
        pulse &= 1          # Force all bits to zero except the first one
        
        # The pulse generator arrays is made up of 32 bit words
        # Here we split the array index into a memory address and a bit-index [0 - 31]
        if which < 2:
            position = setIndex/32
            shift = setIndex%32
            self.wb.SetBitsMulti(self._GetChannelRegName(channel, "set_mem"), ((1<<shift),), ((pulse<<shift),), start=position, to=position+1)

        if which != 1:
            position = resIndex/32
            shift = resIndex%32
            self.wb.SetBitsMulti(self._GetChannelRegName(channel, "res_mem"), ((1<<shift),), ((pulse<<shift),), start=position, to=position+1)

    def ResetBadState(self, channel):
        """ robust reset function to get out of the bad `state 1` """
        self.DisableChannel( channel )                  #Outputs Hi-Z
        self.StopChannel( channel )                     #Stop clock
        self.ClearMemory( channel, "*" )                #Clear the pulse generator arrays
        # Define a single SET pulse per cycle, 
        # after two cycles, the MC100EP140 should be in the good
        # state number 3 (output high)
        self._ConfigurePulse(channel, 10, 50, 1, 1)
        self.StartChannel( channel )                    #Enable clock
        time.sleep( 1e-3 )
        self.StopChannel( channel )
        self.ClearMemory( channel, "*" )

    def CreateSinglePulse(self, channel, start, width, polarity = 1):
        """
        channel = 1 | 2
        start, stop in [ns]
        polarity = 0 - negative polarity | 1 - positive polarity
        """
        self.ClearMemory(channel, "*")
        clockPeriod = self.GetClockPeriod()
        
        stop = start + width
        
        if polarity == 0:   # Swap `start` and `stop` for inverted mode
            tmp = start
            start = stop
            stop = tmp
        
        startBit = int(start/clockPeriod)
        startFineDelay = start - startBit * clockPeriod
        
        stopBit = int(stop/clockPeriod)
        stopFineDelay = stop - stopBit * clockPeriod

        self.SetFineDelay(channel, "set", startFineDelay)
        self.SetFineDelay(channel, "res", stopFineDelay)
        
        self._ConfigurePulse(channel, startBit, stopBit, 1)
    
    def CreateBunchPulses(self, channel, pulseDelay, pulseWidth, bunches, polarity = 1):
        """
        create set of pulses inside LHC bunches
        
        channel = 1 | 2
        pulseDelay, pulseWidth in [ns] - shape of the pulse inside the LHC bunch (25 ns)
        bunches - list of LHC 25 ns bunches, where pulse should be generated
        polarity = 0 - negative polarity | 1 - positive polarity
        """
        self.ClearMemory(channel, "*")
        clockPeriod = self.GetClockPeriod()
        lhcBunchPeriod = 24.95

        assert (pulseDelay >= 0) and (pulseDelay < lhcBunchPeriod), "Pulse starts out of the LHC bunch period!"
        if (pulseWidth > lhcBunchPeriod):
            warnings.warn("Pulse width cannot be > the LHC bunch period! LHC bunch period will be used for a pulse width.", stacklevel=2)
            pulseWidth = lhcBunchPeriod
        pulseStop = pulseDelay + pulseWidth
        assert (pulseStop > pulseDelay), "Pulse starts after its end!"
        
        if polarity == 0:   # Swap `pulseDelay` and `pulseStop` for inverted mode
            tmp = pulseDelay
            pulseDelay = pulseStop
            pulseStop = tmp
        
        # calculate pulse shape inside the bunch
        rfToBunchRatio = int(round(lhcBunchPeriod / clockPeriod))
        if rfToBunchRatio != 5:
            warnings.warn("rfToBunchRatio is not 5 - it should be, in the 'normal' setup!", stacklevel=2)

        pulseStartBit = int(pulseDelay/clockPeriod)
        pulseStartFineDelay = pulseDelay - pulseStartBit * clockPeriod
        
        pulseStopBit = int(pulseStop/clockPeriod)
        pulseStopFineDelay = pulseStop - pulseStopBit * clockPeriod
        
        # set the fine delay
        self.SetFineDelay(channel, "set", pulseStartFineDelay)
        self.SetFineDelay(channel, "res", pulseStopFineDelay)
        
        # configure pulse positions
        if pulseWidth < lhcBunchPeriod:
            # normal pulse configuration
            for bunch in bunches:
                self._ConfigurePulse(channel, (bunch * rfToBunchRatio)+pulseStartBit, (bunch * rfToBunchRatio)+pulseStopBit, 1)
        else:
            # special check for continuous pulses
            for bunch in bunches:
                print "Setting bunch "+str(bunch)
                if (polarity == 1 and (bunch-1) not in bunches) or (polarity == 0 and (bunch+1) not in bunches):
                    self._ConfigurePulse(channel, (bunch * rfToBunchRatio)+pulseStartBit, (bunch * rfToBunchRatio)+pulseStopBit, 1, 1) # only one set bit
                if (polarity == 1 and (bunch+1) not in bunches) or (polarity == 0 and (bunch-1) not in bunches):
                    self._ConfigurePulse(channel, (bunch * rfToBunchRatio)+pulseStartBit, (bunch * rfToBunchRatio)+pulseStopBit, 1, 2) # only one reset bit
        
        # if there are no bunches configured, just maintain the right output level
        if len(bunches) == 0:
            if polarity == 1:
                self._ConfigurePulse(channel, 0, 0, 1, 2) # only one reset bit
            else:
                self._ConfigurePulse(channel, 0, 0, 1, 1) # only one set bit

    def SetClockFineDelay(self, fineDelay, current = 0, capacitors = 0):
        """
        enable and set AD9512 OUT4 output fine delay
        fineDelay = <0, 31>
        """
        # enable fine delay
        self.wb.SetBits("control", (1<<10))
        # set fine delay
        fineDelayReg  = fineDelay & 0x1F
        fineDelayReg |= (current & 0x7) << 5
        fineDelayReg |= (capacitors & 0x7) << 8
        self.wb.Write("fine_delay", fineDelayReg)
    
    def DisableClockFineDelay(self):
        self.wb.SetBits("control", (1<<10), 0)

    def LedModeTest(self):
        self.wb.SetBits("control", (1<<8))

    def LedModeNormal(self):
        self.wb.SetBits("control", (1<<8), 0)
