##-----------------------------------------------------------------------------
## Title      : Python drive: OneWire library
## Project    : FMC DEL 1ns 2cha (FFPG)
## URL        : http://www.ohwr.org/projects/fmc-del-1ns-2cha
##-----------------------------------------------------------------------------
## File       : OneWire.py
## Author(s)  : Jan Pospisil <j.pospisil@cern.ch>
## Company    : CERN (BE-BI-QP)
## Created    : 2016-09-06
## Last update: 2016-09-06
## Standard   : Python
##-----------------------------------------------------------------------------
## Description: OneWire access library for xwb_onewire_master from
##              (http://www.ohwr.org/projects/general-cores/)
##-----------------------------------------------------------------------------
## Copyright (c) 2016 CERN (BE-BI-QP)
##-----------------------------------------------------------------------------
## GNU LESSER GENERAL PUBLIC LICENSE
##-----------------------------------------------------------------------------
## This source file is free software; you can redistribute it and/or modify it
## under the terms of the GNU Lesser General Public License as published by the
## Free Software Foundation; either version 2.1 of the License, or (at your
## option) any later version. This source is distributed in the hope that it
## will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
## of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU Lesser General Public License for more details. You should have
## received a copy of the GNU Lesser General Public License along with this
## source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
##-----------------------------------------------------------------------------
## Revisions  :
## Date        Version  Author        Comment
## 2016-09-06  1.0      Jan Pospisil  Derived from the initial example driver
##-----------------------------------------------------------------------------

import time
from WbOverlay import *

class OneWire(object):
    wb = None
    overdrive = False
    debug = False

    def __init__(self, fmcSlot = -1, wbOverlay = None, overdrive = False):
        """
        fmcSlot = 0 | 1
        wbOverlay = instance of WbOverlay class
        overdrive = 0 - normal mode | 1 - overdrive mode
        """
        if fmcSlot != -1:
            assert (fmcSlot&1) == fmcSlot, "fmcSlot can be just 0 or 1!"
            self.wb = WbOverlay("fmc_fpg", fmcSlot)
        elif (wbOverlay is not None) and (type(wbOverlay) is WbOverlay):
            self.wb = wbOverlay
        else:
            raise Exception("Either fmcSlot or wbOverlay has to be specified!")
        self.overdrive = overdrive and True
    
    def _debug(self, text):
        if (self.debug):
            print(text)
    
    def Reset(self):
        self._debug("OW: reset")
        ctrl = 0x08 | 0x02
        if self.overdrive:
            ctrl |= 4
        self.wb.Write("onewire_ctrl_sta", ctrl)
        reg = self.wb.Read("onewire_ctrl_sta")
        while (0x08 & reg):
            time.sleep( 1e-2 )
            reg = self.wb.Read("onewire_ctrl_sta")
        presence = not (reg&1)
        self._debug("OW: presence = "+str(presence))
        return presence

    def Slot(self, bit):
        bit &= 1
        ctrl = 0x08 | bit
        if self.overdrive:
            ctrl |= 4
        self.wb.Write("onewire_ctrl_sta", ctrl)
        reg = self.wb.Read("onewire_ctrl_sta")
        while (0x08 & reg):
            time.sleep( 1e-2 )
            reg = self.wb.Read("onewire_ctrl_sta")
        return reg & 0x01

    def ReadBit(self):
        result = self.Slot(1)
        self._debug("OW: read bit = "+str(result))
        return result

    def WriteBit(self, bit):
        self._debug("OW: write bit("+str(bit)+")")
        self.Slot(bit)

    def Read(self):
        result = 0;
        for i in range(8):
            result |= (self.ReadBit() << i)
        self._debug("OW: read = "+hex(result))
        return result

    def Write(self, byte):
        self._debug("OW: write("+hex(byte)+")")
        for i in range(8):
            self.WriteBit(byte & 1)
            byte >>= 1
    
    def ReadArray(self, size):
        result = []
        for i in range(size):
            result.append(self.Read())
        return result
