##-----------------------------------------------------------------------------
## Title      : Python drive: WishBone access overlay
## Project    : FMC DEL 1ns 2cha (FFPG)
## URL        : http://www.ohwr.org/projects/fmc-del-1ns-2cha
##-----------------------------------------------------------------------------
## File       : WbOverlay.py
## Author(s)  : Jan Pospisil <j.pospisil@cern.ch>
## Company    : CERN (BE-BI-QP)
## Created    : 2016-09-06
## Last update: 2016-09-06
## Standard   : Python
##-----------------------------------------------------------------------------
## Description: WishBone bus access overlay
##-----------------------------------------------------------------------------
## Copyright (c) 2016 CERN (BE-BI-QP)
##-----------------------------------------------------------------------------
## GNU LESSER GENERAL PUBLIC LICENSE
##-----------------------------------------------------------------------------
## This source file is free software; you can redistribute it and/or modify it
## under the terms of the GNU Lesser General Public License as published by the
## Free Software Foundation; either version 2.1 of the License, or (at your
## option) any later version. This source is distributed in the hope that it
## will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
## of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## See the GNU Lesser General Public License for more details. You should have
## received a copy of the GNU Lesser General Public License along with this
## source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
##-----------------------------------------------------------------------------
## Revisions  :
## Date        Version  Author        Comment
## 2016-09-06  1.0      Jan Pospisil  Derived from the initial example driver
##-----------------------------------------------------------------------------

import sys
sys.path.append("/usr/local/encore")
import encoreio
from CommonFunctions import *

class WbOverlay(object):

    # TODO: add Offset everywhere (also to debug)

    module = None # encoreio.Module
    debug = False # output low-level transactions
    debugSv = False # write SystemVerilog file with all low-level transactions
    debugSvFp = None # 'file' object of SystemVerilog debugging output file
    
    def __init__(self, drvName, lun):
        self.module = encoreio.Module.get_instance(drvName, lun)
    
    def __del__(self):
        self.SetDebugSv("")
    
    def SetDebug(self, debug):
        self.debug = (debug == True)
    
    def PrintDebug(self, text):
        if (self.debug):
            print(text)

    def PrintDebugHex(self, data):
        if (self.debug):
            hexDump(data)
        
    def SetDebugSv(self, filePath):
        """
        set actual SystemVerilog debug output to 'filePath'
        if 'filePath' is "", SystemVerilog debug is disabled
        any actually enabled SystemVerilog debug is closed when this function is called
        """
        if self.debugSv:
            # close SystemVerilog debugging stream
            self.debugSvFp.write("\n")
            self.debugSvFp.write("$display(\"Error count: %0d\\n\", errors);\n")
            self.debugSvFp.close()
            self.debugSv = False
        if filePath <> "":
            # open SystemVerilog debugging stream
            self.debugSvFp = open(filePath, "w")
            self.debugSvFp.write("uint32_t dataOut;\n")
            self.debugSvFp.write("int errors = 0;\n")
            self.debugSvFp.write("\n")
            self.debugSv = True
    
    def _GetRegisterByteAddress(self, RegisterName, Offset = 0):
        Register = self.module[RegisterName]
        RegisterAddress = int(Register["offset"], 16)
        RegisterWordSize = encoreio._Register.wsize_to_dwidth[Register['wsize']]
        return RegisterAddress + Offset*RegisterWordSize/8

    def Read(self, register, Offset = 0):
        result = self.module.read(register, start = Offset, to = Offset+1)[0]
        RegNameVerbose = register
        if Offset > 0:
            RegNameVerbose = RegNameVerbose + " + 0x%x" % Offset
        self.PrintDebug(ConColors.MAGENTA+"WB_READ:"+ConColors.NONE+" ["+RegNameVerbose+"] = "+hex(result))
        if self.debugSv:
            address = self._GetRegisterByteAddress(register, Offset)
            self.debugSvFp.write("WbRead(32'h%08x, dataOut, \"WB_READ: %s:\");\n" % (address, RegNameVerbose))
            self.debugSvFp.write("assert (dataOut == 32'h%08x) else begin\n" % result)
            self.debugSvFp.write("    $error(\"WB_READ different, result should be "+hex(int(result))+"!\");\n")
            self.debugSvFp.write("    errors = errors + 1;\n")
            self.debugSvFp.write("end\n")
        return result

    def Write(self, register, data, Offset = 0):
        RegNameVerbose = register
        if Offset > 0:
            RegNameVerbose = RegNameVerbose + " + 0x%x" % Offset
        self.PrintDebug(ConColors.CYAN+"WB_WRITE:"+ConColors.NONE+" ["+RegNameVerbose+"] <- "+hex(data))
        if self.debugSv:
            address = self._GetRegisterByteAddress(register, Offset)
            self.debugSvFp.write("WbWrite(32'h%08x, 32'h%08x, \"WB_WRITE: %s:\");\n" % (address, data, RegNameVerbose))
        self.module.write(register, (data,), start = Offset, to = Offset+1)

    def SetBits(self, register, mask, bits = -1):
        if bits == -1:
            bits = mask
        value = self.Read(register)
        value &= (~mask)
        value |= (bits & mask)
        self.Write(register, value)

    def ReadMulti(self, register, start = 0, to = -1):
        result = self.module.read(register, start=start, to=to)
        self.PrintDebug(ConColors.MAGENTA+"WB_READ_MULTI:"+ConColors.NONE+" ["+register+"]("+str(start)+":"+str(to)+") = ")
        self.PrintDebugHex( result )
        if self.debugSv:
            address = int(self.module[register]["offset"], 16)
            if to == -1:
                to = int(self.module[register]["depth"], 16)
            for i in range(start, to):
                self.debugSvFp.write("WbRead("+str(address+i*4)+", dataOut, \"WB_READ_MULTI: "+register+"["+str(i)+"]:\");\n")
                self.debugSvFp.write("assert (dataOut == "+str(result[i-start])+") else begin\n")
                self.debugSvFp.write("    $error(\"WB_READ_MULTI different, result should be "+hex(int(result[i-start]))+"!\");\n")
                self.debugSvFp.write("    errors = errors + 1;\n")
                self.debugSvFp.write("end\n")
        return result

    def WriteMulti(self, register, data, start = 0, to = -1):
        self.PrintDebug(ConColors.CYAN+"WB_WRITE_MULTI:"+ConColors.NONE+" ["+register+"]("+str(start)+":"+str(to)+") <- ")
        self.PrintDebugHex( data )
        if self.debugSv:
            address = int(self.module[register]["offset"], 16)
            if to == -1:
                to = int(self.module[register]["depth"])
            for i in range(start, to):
                self.debugSvFp.write("WbWrite("+str(address+i*4)+", "+str(data[i-start])+", \"WB_WRITE_MULTI: "+register+"["+str(i)+"]:\");\n")
        self.module.write(register, data, start=start, to=to)

    def SetBitsMulti(self, register, mask, bits, start = 0, to = -1):
        assert len(mask) == len(bits), "Different MASK and BITS length!"
        values = self.ReadMulti(register, start, to)
        values = list(values)
        assert len(values) == len(mask), "Different length of MASK/BITS and actual data!"
        for i in range(len(values)):
            values[i] &= (~mask[i])
            values[i] |= (bits[i] & mask[i])
        self.WriteMulti(register, values, start, to)

    def WaitForNotBusy(self, register, bit, timeOut, tries, message = "", polarity = 1):
        """
        timeOut in [s] (floating point)
        polarity = 1: (register.bit == 1) <=> BUSY, wait for 0
        polarity = 0: (register.bit == 0) <=> BUSY, wait for 1
        """
        busy = True
        for i in range(tries):
            if (self.Read(register) & (1<<bit)) != polarity:
                busy = False
                break
            # wait for timeOut seconds
            timeToWait = timeOut
            while timeToWait > 0:
                timeBegin = time.time()
                time.sleep(timeToWait)
                timeToWait -= time.time() - timeBegin
        if busy:
            raise Exception("WaitForNotBusy error: "+message)
